<div class="overflow-x-auto overflow-y-clip h-max">
    <table class="table-auto w-full text-white">
        {% for issue in issues %}
        <tbody class="first:border-t border-b border-zinc-500">
            {% include "scan_report/html/_issue.html" %}
        </tbody>
        {% endfor %}
    </table>
</div>