<div class="p-4">
  <h2 class="my-4 mr-2 font-medium">What's next?</h2>
  <div class="text-sm my-2">
    <h3 class="text-sm my-1">
      1. Generate a test suite from your scan results
    </h3>
    <pre><code class="language-python rounded">{% raw %}test_suite = results.generate_test_suite("My first test suite"){% endraw %}</code></pre>
    <br />
    <h3 class="text-sm my-1">2. Run your test suite</h3>
    <pre><code class="language-python rounded">{% raw %}test_suite.run(){% endraw %}</code></pre>
  </div>
</div>
