{% block head %}
<style>
    .alert {
        padding: 16px;
        border-radius: 4px;
        display: flex;
        align-items: center;
        flex-wrap: wrap;
        gap: 4px;
        margin-bottom: 16px;
    }

    .test-result {
        border-radius: 4px;
        padding: 0 12px;
        display: flex;
        gap: 4px;
        align-items: center;
    }

    .alert-error, .test-card-failed .test-result {
        color: rgb(255, 82, 82);
        background: rgb(243, 226, 226);
        fill: currentcolor;
    }

    .test-card-error .test-result {
        color: #856404;
        background: #fff3cd;
        fill: currentcolor;
    }

    .alert-success, .test-card-passed .test-result {
        color: rgb(76, 175, 80);
        background: rgb(226, 243, 226);
        fill: currentcolor;
    }

    .alert svg {
        height: 24px;
        width: 24px;
        border-radius: 50%;
        background: rgba(0, 0, 0, 10%);
    }

    .test-result svg {
        height: 16px;
        width: 16px;
    }

    .card-container {
        display: flex;
        flex-direction: column;
        gap: 16px;
    }

    .test-card {
        border-radius: 4px;
        border: 1px solid #e0e0e0;
        color: rgb(98, 98, 98);
        background: #fff;
    }

    .param {
        display: flex;
        flex-direction: column;
    }

    .test-card .test-name, .test-card .param-value {
        color: #000
    }

    .test-card-row {
        padding: 10px;
        display: flex;
        gap: 20px;
        align-items: center;
        flex-wrap: wrap;
    }

    .test-card-row:last-child {
        border-top: 1px solid #dee2e6;
    }

    .spacer {
        flex-grow: 1;
    }

    .test-card-failed .metric {
        color: #b71c1c;
    }
</style>

{% endblock %}

{% block content %}
<div class='dark'>
    <div id='gsk-suite'>
        {% include "suite_results/_suite_results_header.html" %}
        {% include "suite_results/_suite_results_cards.html" %}
    </div>
</div>
{% endblock %}
