body{
  background: #18181B;
}

.main{
  font-family: "Noto Sans", ui-sans-serif, system-ui, -apple-system, BlinkMacSystemFont, "Segoe UI", Roboto, "Helvetica Neue", Arial, sans-serif, "Apple Color Emoji", "Segoe UI Emoji", "Segoe UI Symbol", "Noto Color Emoji";
  color: #FDFDFD;
}

h1 {
  font-size: 2.5rem;
  color: white;
}

h3 {
  font-size: 1.5rem;
  background: #0c087c;
  padding: 10px;
  margin: 0px;
  border: 1px solid #6b7280;}

.extended-title{
  width:100%;
}

#gsk-overview{
  display:flex;
}

h4 {
  font-size: 1rem;
  background: #27272A;
  padding: 10px;
  margin: 0px;
  border-bottom: 1px solid #6b7280;
}

h2 {
  font-size: 1.5rem;
  margin-top: 3px;
  color:#000000;
}

.header{
  display: flex;
  justify-content: center;
  align-items: center;
}
.header > * {
  margin-inline: 20px;
}

.flex-row {
  display: flex;
  flex-direction: row;
  padding:10px;
  border: 1px solid #27272A;
}

.flex-row>div {
  flex: auto;
  box-sizing: border-box;
  display: flex;
  flex-direction: column;
  justify-content: center;
  align-items: center;
}

progress[value] {
  --background: #6D6D6D;
  -webkit-appearance: none;
  -moz-appearance: none;
  appearance: none;
  border: none;
  height: 4px;
  margin: 0 10px;
  border-radius: 10em;
  background: var(--background);
}

progress[value]::-webkit-progress-bar {
  border-radius: 10em;
  background: var(--background);
}

progress[value]::-webkit-progress-value {
  border-radius: 10em;
  background: var(--color);
}

progress[value]::-moz-progress-bar {
  border-radius: 10em;
  background: var(--color);
}

label {
  font-size: 20px;
  font-weight: bold;
  display: block;
  margin: 20px 0;
}

.tab {
  overflow: hidden;
  border: 1px solid #27272A;
  background-color: #27272A;
}

.tab button {
  background-color: inherit;
  float: left;
  border: none;
  outline: none;
  cursor: pointer;
  padding: 14px 16px;
  transition: 0.3s;
  color: #ffffff;
  font-size:1.2rem;
}

.tab div {
  background-color: inherit;
  float: left;
  border: none;
  outline: none;
  cursor: pointer;
  padding: 14px 16px;
  transition: 0.3s;
  color: #ffffff;
  font-size: 1.2rem;
}

.tab button:hover {
  background-color: #18181B;
}

.tab-title{
  font-size: 1.5rem;
  font-weight: bold;
  margin-bottom:-5px;
}

.tab button.active {
  background-color: #18181B;
  border-top: 1px solid #6b7280;
  border-bottom: 1px solid #18181B;
  border-left: 1px solid #6b7280;
  border-right: 1px solid #6b7280;
}

.tabcontent {
  display: none;
  padding: 6px 12px;
  background: #18181B;
  border: 1px solid #27272A;
  border-top: 1px solid #6b7280;
  margin-top: -2px;
}

#gsk-advice {
  display: flex;
  justify-content: center;
}

#gsk-metrics{
  width:100%;
}

#recommendation {
  margin-top: 20px;
  padding: 20px;
  border-radius: 10px;
  background-color: #e1ce86;
  color: #27272A;
  width:95%;
  box-shadow: 0 4px 8px 0 #000000, 0 6px 20px 0 #000000;
  font-size: 12pt;
}

.separator {
  margin: 20px 0;
}

.separator-border {
  margin: 20px 0;
  border-bottom: 1px solid #6b7280;
}

#gsk-rag{
  margin: 32px 28px;
  padding: 12px 24px;
  background-color: #111516;
}

.section-container {
  margin-bottom: 32px;
}

  .components-container {
    display: flex;
    flex-flow: row wrap;
    align-items: flex-start;
    gap: 0 32px;
  }

    .component-card {
      background-color: #14191B;
      border-radius: 16px;
      padding: 28px 32px 32px 32px;
      display: flex;
      flex-flow: column;
      align-items: center;
      margin-top: 32px;
      flex-grow: 1;
    }

    .component-title {
      font-size: 12px;
      font-weight: 500;
      color: #B1B1B1;
      padding-bottom: 8px;
    }

    .component-value {
      font-size: 32px;
      font-weight: 500;
      padding-bottom: 12px;
    }
      
      .text-green {
        color: #04B543;
      }
      
      .text-orange {
        color: #E76E0F;
      }
      
      .text-red {
        color: #EA3829;
      }

      .tooltip {
        position: relative;
        display: inline-block;
      }

      .tooltip .tooltiptext {
        visibility: hidden;
        width: 120px;
        background-color: #464646;
        color: #E6E6E6;
        text-align: center;
        border-radius: 6px;
        position: absolute;
        z-index: 1;
        top: 150%;
        left: 50%;
        margin-left: -60px;
        font-size: 12px;
        padding: 12px;
      }
      
      .tooltip .tooltiptext::after {
        content: "";
        position: absolute;
        bottom: 100%;
        left: 50%;
        margin-left: -5px;
        border-width: 5px;
        border-style: solid;
        border-color: transparent transparent black transparent;
      }
      
      .tooltip:hover .tooltiptext {
        visibility: visible;
      }

    .overall-card {
      background-color: #026836;
      border-radius: 16px;
      padding: 28px 32px 32px 32px;
      display: flex;
      flex-flow: column;
      align-items: center;
      justify-content: center;
      margin-top: 32px;
      flex-grow: 1;
    }
  
    .overall-title {
      font-size: 12px;
      font-weight: 500;
      color: #E6E6E6;
      padding: 14px 0 8px 0;
      text-transform: uppercase;
     }
  
    .overall-value {
      font-size: 32px;
      font-weight: 500;
      padding-bottom: 12px;
      color: #E6E6E6;
    }

.section-title {
  font-size: 12px;
  color: #B1B1B1;
  margin-bottom: 20px;
  text-align: left;
  width: 100%;
}

.section-content {
  color: #E6E6E6;
  font-size: 20px;
  line-height: 1.5;
}

.section-card {
  background-color: #14191B;
  border-radius: 16px;
  padding: 28px 32px 32px 32px;
  display: flex;
  flex-flow: column;
  align-items: center;
}

.correctness-indicator{
  padding: 20px;
  border-radius: 50px;
  font-size: 16pt;
  box-shadow: 0 4px 8px 0 #000000, 0 6px 20px 0 #000000;
  text-align: center;
}

.metric-title{
  margin: -2px;
  border-bottom: none;
}

.hist-row {
  display: flex;
  flex-direction: row;
  padding: 10px;
  justify-content: space-around;
  width: 85%;
}

.hist-row>div {
  flex: auto;
  box-sizing: border-box;
  display: flex;
  flex-direction: column;
  justify-content: center;
  align-items: center;
  padding-left: 1%;
  padding-right: 1%;
}

.tab-row{
  display: flex;
  flex-direction: column;
  align-items: center;
}

#component-table{
  width:50%;
  margin-top: 10px;
}

.green{
  background-color: #0a980a;
}

.orange {
  background-color: #e5b62a;
}

.red {
  background-color: #ba0e0e;
}

.progress-green {
  --color: #04B543;
}

.progress-orange {
  --color: #E76E0F;
}

.progress-red {
  --color: #EA3829;
}

.corr-plot{
  flex: 1;
  padding-left: 2%;
}

.tooltip-text {
  position: absolute;
  display: none;
  visibility: hidden;
  z-index: 1;
  top: 100%;
  left: 0%;
  width: 100%;
  color: white;
  font-size: 12px;
  background-color: #2d3d4c;
  border-radius: 10px;
  padding: 10px 15px 10px 15px;
}

#fade {
  opacity: 1;
  transition: opacity 0.5s;
}

#delay {
  opacity: 0;
  transition: opacity 0.2s;
  transition-delay: 1s;
}

td {
  position: relative;
}

tr:hover .tooltip-text {
  display: block;
  visibility: visible;
}

.tr:hover #fade {
  opacity: 1;
}

.tr:hover #delay {
  opacity: 1;
}

.callout {
  padding: 0.5rem 1rem 0.5rem 3rem;
  background: #D9EDF9;
  border: 3px solid #0088D1;
  color: #272eb5;
  position: relative;
  max-width: 40rem;
  border-radius: 10px;
  margin-top: 10%;
  font-size: 11pt;
}

.callout-icon {
  content: "";

  /* SVG via a data URI! */
  background-size: cover;
  width: 1.5rem;
  height: 1.5rem;
  display: block;
  position: absolute;
  left: 0.9rem;
  top: 1.1rem;
}

.callout-icon svg{
  fill: #016ca7;
}
.callout p+p {
  margin-top: 1em;
}

.callout a {
  color: #272eb5;
  font-weight: bold;
}

#gsk-logo {
  padding-top: 10px;
}
